@@ -1,5 +1,6 @@
 import json
 import logging
+from .utils import is_invalid_params
 from .exceptions import (
     JSONRPCInvalidParams,
     JSONRPCInvalidRequest,
@@ -41,24 +42,24 @@ class JSONRPCResponseManager(object):
     }
 
     @classmethod
-    def handle(cls, request_str, dispatcher):
+    def handle(cls, request_str, dispatcher, context=None):
         if isinstance(request_str, bytes):
             request_str = request_str.decode("utf-8")
 
         try:
-            json.loads(request_str)
+            data = json.loads(request_str)
         except (TypeError, ValueError):
             return JSONRPC20Response(error=JSONRPCParseError()._data)
 
         try:
-            request = JSONRPCRequest.from_json(request_str)
+            request = JSONRPCRequest.from_data(data)
         except JSONRPCInvalidRequestException:
             return JSONRPC20Response(error=JSONRPCInvalidRequest()._data)
 
-        return cls.handle_request(request, dispatcher)
+        return cls.handle_request(request, dispatcher, context)
 
     @classmethod
-    def handle_request(cls, request, dispatcher):
+    def handle_request(cls, request, dispatcher, context=None):
         """ Handle request data.
 
         At this moment request has correct jsonrpc format.
@@ -71,7 +72,7 @@ class JSONRPCResponseManager(object):
         """
         rs = request if isinstance(request, JSONRPC20BatchRequest) \
             else [request]
-        responses = [r for r in cls._get_responses(rs, dispatcher)
+        responses = [r for r in cls._get_responses(rs, dispatcher, context)
                      if r is not None]
 
         # notifications
@@ -79,44 +80,64 @@ class JSONRPCResponseManager(object):
             return
 
         if isinstance(request, JSONRPC20BatchRequest):
-            return JSONRPC20BatchResponse(*responses)
+            response = JSONRPC20BatchResponse(*responses)
+            response.request = request
+            return response
         else:
             return responses[0]
 
     @classmethod
-    def _get_responses(cls, requests, dispatcher):
+    def _get_responses(cls, requests, dispatcher, context=None):
         """ Response to each single JSON-RPC Request.
 
         :return iterator(JSONRPC20Response):
 
+        .. versionadded: 1.9.0
+          TypeError inside the function is distinguished from Invalid Params.
+
         """
         for request in requests:
-            def response(**kwargs):
-                return cls.RESPONSE_CLASS_MAP[request.JSONRPC_VERSION](
+            def make_response(**kwargs):
+                response = cls.RESPONSE_CLASS_MAP[request.JSONRPC_VERSION](
                     _id=request._id, **kwargs)
+                response.request = request
+                return response
 
+            output = None
             try:
                 method = dispatcher[request.method]
             except KeyError:
-                output = response(error=JSONRPCMethodNotFound()._data)
+                output = make_response(error=JSONRPCMethodNotFound()._data)
             else:
                 try:
-                    result = method(*request.args, **request.kwargs)
-                except TypeError:
-                    output = response(error=JSONRPCInvalidParams()._data)
+                    kwargs = request.kwargs
+                    if context is not None:
+                        context_arg = dispatcher.context_arg_for_method.get(
+                            request.method)
+                        if context_arg:
+                            context["request"] = request
+                            kwargs[context_arg] = context
+                    result = method(*request.args, **kwargs)
                 except JSONRPCDispatchException as e:
-                    output = response(error=e.error._data)
+                    output = make_response(error=e.error._data)
                 except Exception as e:
                     data = {
                         "type": e.__class__.__name__,
                         "args": e.args,
                         "message": str(e),
                     }
+
                     logger.exception("API Exception: {0}".format(data))
-                    output = response(
-                        error=JSONRPCServerError(data=data)._data)
+
+                    if isinstance(e, TypeError) and is_invalid_params(
+                            method, *request.args, **request.kwargs):
+                        output = make_response(
+                            error=JSONRPCInvalidParams(data=data)._data)
+                    else:
+                        output = make_response(
+                            error=JSONRPCServerError(data=data)._data)
                 else:
-                    output = response(result=result)
+                    output = make_response(result=result)
             finally:
                 if not request.is_notification:
                     yield output
